from .quantity import Quantity
from .temperature import Temperature
from .energy import Energy


class GasConstant(Quantity):

    @property
    def si(self):
        return 8.314

    @property
    def JmolK(self):
        return self.si

    @property
    def m3PaKmol(self):
        return self.si

    @property
    def kJmolK(self):
        return 0.008314

    @property
    def LbarKmol(self):
        return 0.08314
    
    def __mul__(self, other):
        if isinstance(other, Temperature):
            si = self.si * other.si
            return Energy(si=si)
        return super().__mul__(other)

    def __rmul__(self, other):
        if isinstance(other, Temperature):
            si = other.si * self.si 
            return Energy(si=si)
        return super().__rmul__(other)



import plotly.graph_objects as go, numpy as np


def plot(config):
    fig = go.Figure()
    for key, value in config["traces"].items():
        x = value.get("x")
        y = value["y"]
        if x is not None:
            xmin = np.min(x)
            xmax = np.max(x)
        if isinstance(y, (float, int)):
            x = [xmin, xmax]
            y = [y, y]
        trace = go.Scatter(
            x=x,
            y=y,
            mode=value["mode"],
            name=key,
            text=value.get("hover_labels"),
            hovertemplate="<b>%{text}</b><br>X: %{x}<br>Y: %{y}<extra></extra>",
        )
        fig.add_trace(trace)

    xmin = config.get("xmin")
    xmax = config.get("xmax")
    ymin = config.get("ymin")
    ymax = config.get("ymax")

    fig.update_layout(
        title=dict(text=f"<b>{config['title']}</b>", x=0.5),
        xaxis_title=f"<b>{config['xlabel']}</b>",
        yaxis_title=f"<b>{config['ylabel']}</b>",
        width=650,
        margin=dict(t=50, b=110),
        yaxis=dict(range=[ymin, ymax]),
        xaxis=dict(range=[xmin, xmax]),
        annotations=config.get('info_text'),
        # colorway=px.colors.qualitative.D3
        # colorway=px.colors.qualitative.Set1
        # colorway=px.colors.qualitative.Pastel
        # colorway=px.colors.qualitative.Dark2
    )

    fig.show()



from .quantity import Quantity
from catrxneng.utils import *


class GHSV(Quantity):

    @property
    def molskgcat(self):
        return self.si

    @molskgcat.setter
    def molskgcat(self, value):
        self.si = to_float(value)

    @property
    def smLhgcat(self):
        return self.si * 3600 * 22.4

    @smLhgcat.setter
    def smLhgcat(self, value):
        self.si = to_float(value) / 3600 / 22.4

    def __mul__(self, other):
        from .mass import Mass
        from .molar_flow_rate import MolarFlowRate

        if isinstance(other, Mass):
            si = self.si * other.si
            return MolarFlowRate(si=si)
        return super().__mul__(other)

    def __rmul__(self, other):
        from .mass import Mass
        from .molar_flow_rate import MolarFlowRate

        if isinstance(other, Mass):
            si = self.si * other.si
            return MolarFlowRate(si=si)
        return super().__rmul__(other)


from .species import Species
from ..quantities import *


class CO(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=298)
        self.max_temp = Temperature(K=1300)
        self.Hf298 = Energy(kJmol=-110.53)
        self.thermo_params = {
            "A": 25.56759,
            "B": 6.096130,
            "C": 4.054656,
            "D": -2.671301,
            "E": 0.131021,
            "F": -118.0089,
            "G": 227.3665,
            "H": -110.5271,
        }
        super().__init__(T, stoich_coeff)



from .species import Species
from ..quantities import *

class H2(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=298)
        self.max_temp = Temperature(K=1000)
        self.Hf298 =  Energy(kJmol=0)
        self.thermo_params = {
            "A": 33.066178,
            "B": -11.363417,
            "C": 11.432816,
            "D": -2.772874,
            "E": -0.158558,
            "F": -9.980797,
            "G": 172.707974,
            "H": 0,
        }
        super().__init__(T, stoich_coeff)



from .quantity import Quantity


class MolarGasVolume():
    def __init__(self):
        self.STP = Quantity(Lmol=22.4)
        self.STP.m3mol = self.STP.Lmol / 1000
        self.STP.si = self.STP.m3mol
        self.NTP = Quantity(Lmol=24.05)
        self.NTP.m3mol = self.NTP.Lmol / 1000
        self.NTP.si = self.NTP.m3mol
    




import numpy as np, pandas as pd


def k(T, Ea, k0=None, kref=None, Tref=None):
    from ..quantities import R, RateConstant

    if kref and Tref:
        k0 = kref / (np.exp(-Ea / (R * Tref)))
        k0 = RateConstant(si=k0.si, order=kref.order)

    k = k0 * np.exp(-Ea / (R * T))
    return RateConstant(si=k.si, order=k0.order)


def to_float(value):
    if isinstance(value, np.ndarray):
        return value.astype(float)
    if isinstance(value, int):
        return float(value)
    return value


def plot_info_box(text):
    annotations = [
        dict(
            x=0.5,
            y=-0.32,
            xref="paper",
            yref="paper",
            text=text,
            showarrow=False,
            font=dict(size=12, style="italic", weight="bold"),
        )
    ]
    return annotations





from .quantity import Quantity
from ..utils import *


class Energy(Quantity):

    @property
    def Jmol(self):
        return self.si

    @Jmol.setter
    def Jmol(self, value):
        self.si = to_float(value)

    @property
    def kJmol(self):
        return self.si / 1000

    @kJmol.setter
    def kJmol(self, value):
        self.si = to_float(value * 1000)

    @property
    def kcalmol(self):
        return self.si / 4184

    @kcalmol.setter
    def kcalmol(self, value):
        self.si = to_float(value * 4184)



from .quantity import Quantity
from catrxneng.utils import *


class InvGHSV(Quantity):

    @property
    def molskgcat(self):
        return self.si

    @molskgcat.setter
    def molskgcat(self, value):
        self.si = to_float(value)

    @property
    def smLhgcat(self):
        return self.si / 3600 / 22.4

    @smLhgcat.setter
    def smLhgcat(self, value):
        self.si = to_float(value) * 3600 * 22.4



from .quantity import Quantity
from catrxneng.utils import *


class MolarFlowRate(Quantity):

    @property
    def mols(self):
        return self.si

    @mols.setter
    def mols(self, value):
        self.si = to_float(value)

    @property
    def molmin(self):
        return self.si * 60

    @molmin.setter
    def molmin(self, value):
        self.si = to_float(value / 60)

    @property
    def molh(self):
        return self.si * 3600

    @molh.setter
    def molh(self, value):
        self.si = to_float(value / 3600)

    @property
    def smLmin(self):
        return self.si * 60 * 22.4 * 1000

    @smLmin.setter
    def smLmin(self, value):
        self.si = to_float(value / 60 / 22.4 / 1000)

    @property
    def nmLmin(self):
        return self.si * 60 * 24.05 * 1000

    @nmLmin.setter
    def nmLmin(self, value):
        self.si = to_float(value / 60 / 24.05 / 1000)
    


from .species import Species
from ..quantities import *


class N2(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=500) 
        self.max_temp = Temperature(K=2000)
        self.Hf298 =  Energy(kJmol=0)
        self.thermo_params = {
            "A": 19.50583,
            "B": 19.88705,
            "C": -8.598535,
            "D": 1.369784,
            "E": 0.527601,
            "F": -4.935202,
            "G": 212.3900,
            "H": 0.0,
        }
        super().__init__(T, stoich_coeff)




from .species import Species
from ..quantities import *

class CH4(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=298)
        self.max_temp = Temperature(K=1300)
        self.Hf298 = Energy(kJmol=-74.6)
        self.thermo_params = {
            "A": -0.703029,
            "B": 108.4773,
            "C": -42.52157,
            "D": 5.862788,
            "E": 0.678565,
            "F": -76.84376,
            "G": 158.7163,
            "H": -74.87310
        }
        super().__init__(T, stoich_coeff)



from .species import Species
from ..quantities import *


class NH3(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=298) 
        self.max_temp = Temperature(K=1400)
        self.Hf298 =  Energy(kJmol=-45.9)
        self.thermo_params = {
            "A": 19.99563,
            "B": 49.77119,
            "C": -15.37599,
            "D": 1.921168,
            "E": 0.189174,
            "F": -53.30667,
            "G": 203.8591,
            "H": -45.89806,
        }
        super().__init__(T, stoich_coeff)




from .quantity import Quantity


class RateConstant(Quantity):
    def __init__(self, order=None, **kwargs):
        self.order = order
        super().__init__(**kwargs)

    @property
    def molskgcatPa(self):
        return self.si

    @molskgcatPa.setter
    def molskgcat(self, value):
        from ..utils import to_float
        self.si = to_float(value)

    @property
    def molhgcatbar(self):
        return self.si * 3600 / 1000 * (100000**self.order)

    @molhgcatbar.setter
    def molhgcatbar(self, value):
        from ..utils import to_float
        self.si = to_float(value / 3600 * 1000 / (100000**self.order))



from .species import Species
from ..quantities import *


class H2O(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=500)
        self.max_temp = Temperature(K=1700)
        self.Hf298 = Energy(kJmol=-241.83)
        self.thermo_params = {
            "A": 30.09200,
            "B": 6.832514,
            "C": 6.793435,
            "D": -2.534480,
            "E": 0.082139,
            "F": -250.8810,
            "G": 223.3967,
            "H": -241.8264
        }
        super().__init__(T, stoich_coeff)



import numpy as np

from .reaction import Reaction
from ..quantities import *
from ..species import CO, H2, C2H4, H2O, Ar


class FTS(Reaction):
    def __init__(self, T):
        # self.components = {
        #     "CO": CO(T=T, stoich_coeff=-2),
        #     "H2": H2(T=T, stoich_coeff=-4),
        #     "C2H4": C2H4(T=T, stoich_coeff=1),
        #     "H2O": H2O(T=T, stoich_coeff=2),
        #     "inert": Ar(T=T, stoich_coeff=0)
        # }
        self.components = (CO(T), H2(T), C2H4(T), H2O(T), Ar(T))
        self.stoich_coeff = Unitless(si=[-2., -4., 1., 2., 0.])
        super().__init__(T)



import numpy as np
from scipy.optimize import minimize

from catrxneng.quantities import *


class Reaction:
    def __init__(self, T):
        self.T = T
        self.fug_coeff = Unitless(si=np.ones(len(self.components)))
        self.update()

    @property
    def comp_list(self):
        return (comp.__class__.__name__ for comp in self.components)

    def update(self):
        for comp in self.components:
            comp.T = self.T
        self.Hf = Energy(si=[comp.Hf.si for comp in self.components])
        self.Sf = Entropy(si=[comp.Sf.si for comp in self.components])
        self.dHr = np.sum(self.Hf * self.stoich_coeff)
        self.dSr = np.sum(self.Sf * self.stoich_coeff)
        self.dGr = self.dHr - self.dSr * self.T
        self.Keq = np.exp(-self.dGr / (R * self.T))

    def check_components(self, p0):
        if p0.size != len(self.components):
            raise ValueError(
                "Number of components for reactor and rate model do not match."
            )

    def eq_conv(self, p0):
        self.check_components(p0)
        P = np.sum(p0)
        initial_total_mol = 100
        initial_molfrac = p0 / P
        initial_mol = initial_molfrac * initial_total_mol
        std_state_fugacity = Pressure(atm=np.ones(len(self.components)))

        def objective(extent):

            extent = Quantity(si=extent)
            mol = initial_mol + extent * self.stoich_coeff
            total_mol = np.sum(mol)
            molfrac = mol / total_mol
            fugacity = molfrac * self.fug_coeff * P
            activity = fugacity / std_state_fugacity
            Ka = np.prod(activity**self.stoich_coeff)
            # Kx = np.prod(molfrac**self.stoich_coeff)
            # Kphi = np.prod(self.fug_coeff**self.stoich_coeff)
            # Kp = np.prod(P**self.stoich_coeff)
            # Kf0 = np.prod(std_state_fugacity**self.stoich_coeff)
            return ((Ka - self.Keq) ** 2).si * 1e5

        adj_init_mol_reactants = np.array(
            [
                mol / stoich_coeff
                for mol, stoich_coeff in zip(initial_mol.si, self.stoich_coeff.si)
                if stoich_coeff < 0
            ]
        )
        bounds = [(0.00001, np.min(-adj_init_mol_reactants) * 0.99999)]
        initial_guess = -0.5 * initial_mol.si[0] / self.stoich_coeff.si[0]
        result = minimize(
            objective,
            initial_guess,
            bounds=bounds,
            options={"ftol": 1e-10},
        )
        if result.success:
            extent = result.x[0]
            conv = -self.stoich_coeff.si[0] * extent / initial_mol.si[0]
            return Unitless(si=conv)
        else:
            raise ValueError("Optimization failed: " + result.message)



from .quantity import Quantity
from .energy import Energy
from .temperature import Temperature
from .entropy import Entropy
from .gas_constant import GasConstant
R = GasConstant()
from .pressure import Pressure
from .unitless import Unitless
from .mass import Mass
from .ghsv import GHSV
from .molar_flow_rate import MolarFlowRate
from .reaction_rate import ReactionRate
from .rate_constant import RateConstant
from .vol_flow_rate import VolumetricFlowRate
from .inv_pressure import InversePressure
from .inv_ghsv import InvGHSV


from .quantity import Quantity
from catrxneng.utils import *


class Pressure(Quantity):

    @property
    def Pa(self):
        return self.si

    @Pa.setter
    def Pa(self, value):
        self.si = to_float(value)

    @property
    def bar(self):
        return self.si / 100000

    @bar.setter
    def bar(self, value):
        self.si = to_float(value) * 100000

    @property
    def kPa(self):
        return self.si / 1000 

    @kPa.setter
    def kPa(self, value):
        self.si = to_float(value) * 1000

    @property
    def atm(self):
        return self.si / 101325 

    @atm.setter
    def atm(self, value):
        self.si = to_float(value) * 101325

    @property
    def MPa(self):
        return self.si / 1000000 

    @MPa.setter
    def MPa(self, value):
        self.si = to_float(value) * 1000000


import numpy as np
from scipy.integrate import solve_ivp

from ..quantities import *
from .reactor import Reactor


class PFR(Reactor):
    def __init__(self, conditions, rate_model):
        super().__init__()
        self.conditions = conditions
        self.Ft0 = self.conditions["ghsv"] * self.conditions["mcat"]
        self.conditions["P0"] = np.sum(self.conditions["p0"])
        self.y0 = self.conditions["p0"] / self.conditions["P0"]
        self.F0 = self.y0 * self.Ft0
        self.rate_model = rate_model
        self.check_components()

    def dFdw(self, w, F):
        F = MolarFlowRate(molh=F)
        Ft = np.sum(F)
        y = F / Ft
        p = y * self.conditions["P0"]
        T = self.conditions["T"]
        return np.array([rate(T, p).molhgcat for rate in self.rate_model.rate])

    def solve(self):
        w_span = (0, self.conditions["mcat"].g)
        w_eval = np.linspace(0, self.conditions["mcat"].g, 1000)
        solution = solve_ivp(self.dFdw, w_span, self.F0.molh, t_eval=w_eval)
        self.w = Mass(g=solution.t)
        self.F = MolarFlowRate(molh=solution.y)
        self.Ft = np.sum(self.F, axis=0)
        self.y = self.F / self.Ft
        self.conv = (self.F0[0] - self.F[0]) / self.F0[0]
        self.inv_ghsv = InvGHSV(si=self.w.si / self.Ft0.si)
        self.ghsv = GHSV(si=1/self.inv_ghsv.si)
        vol_flow_rate = self.Ft0 * R * self.conditions["T"] / self.conditions["P0"]
        self.v = VolumetricFlowRate(si=vol_flow_rate.si)
        self.tau_mod = self.w.g / self.v.mLs
    
    def rate(self):
        p = self.y * self.conditions["P0"]
        rates = []
        for ri in self.rate_model.rate:
            rates.append(ri(self.conditions["T"], p))
        return rates




import numpy as np

from ..quantities import *


class Species:

    def __init__(self, T, stoich_coeff=None, fug_coeff=1, par_pres=None):
        self.T = T
        self.stoich_coeff = stoich_coeff
        self.par_pres = par_pres
        self.fug_coeff = fug_coeff
        self.update()

    def update(self):
        self.check_temps()
        t = self.T / 1000
        params = self.thermo_params

        dHf = (
            params["A"] * t
            + params["B"] * t**2 / 2
            + params["C"] * t**3 / 3
            + params["D"] * t**4 / 4
            - params["E"] / t
            + params["F"]
            - params["H"]
        )  # kJ/mol
        self.Hf = Energy(kJmol=dHf.si) + self.Hf298

        Sf = (
            params["A"] * np.log(t)
            + params["B"] * t
            + params["C"] * t**2 / 2
            + params["D"] * t**3 / 3
            - params["E"] / (2 * t**2)
            + params["G"]
        ) / 1000 # kJ/mol/K
        self.Sf = Entropy(kJmolK=Sf.si)

        self.Gf = self.Hf + self.T * self.Sf

    def check_temps(self):
        if isinstance(self.T.si, np.ndarray):
            if self.T.si.min() < self.min_temp.si or self.T.si.max() > self.max_temp.si:
                raise ValueError("Temperature out of range for NIST parameters.")
        else:
            if self.T.si < self.min_temp.si or self.T.si > self.max_temp.si:
                raise ValueError("Temperature out of range for NIST parameters.")



from .species import Species
from ..quantities import *


class CO2(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=298)
        self.max_temp = Temperature(K=1200)
        self.Hf298 = Energy(kJmol=-393.51)
        self.thermo_params = {
            "A": 24.99735,
            "B": 55.18696,
            "C": -33.69137,
            "D": 7.948387,
            "E": -0.136638,
            "F": -403.6075,
            "G": 228.2431,
            "H": -393.5224,
        }
        super().__init__(T, stoich_coeff)



from scipy.optimize import fsolve
from .reaction import Reaction
from ..species import CO, H2, CH4, H2O


class COMethanation(Reaction):
    def __init__(self, T):
        self.T = T
        self.components = {
            "CO": CO(T=T, stoich_coeff=-1),
            "H2": H2(T=T, stoich_coeff=-3),
            "CH4": CH4(T=T, stoich_coeff=1),
            "H2O": H2O(T=T, stoich_coeff=2),
        }




import plotly.graph_objects as go, numpy as np


class Reactor:

    def __init__(self):
        self.conditions = {"ghsv": None, "T": None, "p0": None, "mcat": None}

    def check_components(self):
        if self.conditions["p0"].size != len(self.rate_model.components):
            raise ValueError(
                "Number of components for reactor and rate model do not match."
            )

    def plot(self, x, y, names, modes, xlabel, ylabel, title=None):
        fig = go.Figure()
        for i in range(len(y)):
            trace = go.Scatter(x=x[i], y=[i], mode=modes[i], name=names[i])
            fig.add_trace(trace)
        fig.update_layout(
            title=dict(text=f"<b>{title}</b>", x=0.5),
            xaxis_title=f"<b>{xlabel}</b>",
            yaxis_title=f"<b>{ylabel}</b>",
            width=800,
        )
        fig.show()



from .reaction import Reaction
from .. import species


class WGS(Reaction):
    def __init__(self, T):
        self.components = {
            "CO": species.CO(T=T, stoich_coeff=-1),
            "H2O": species.H2O(T=T, stoich_coeff=-1),
            "CO2": species.CO2(T=T, stoich_coeff=1),
            "H2": species.H2(T=T, stoich_coeff=1),
            "inert": species.Ar(T=T, stoich_coeff=0)
        }
        super().__init__(T=T)


from .quantity import Quantity
from .energy import Energy
from .temperature import Temperature
from catrxneng.utils import *


class Entropy(Quantity):

    @property
    def JmolK(self):
        return self.si

    @JmolK.setter
    def JmolK(self, value):
        self.si = to_float(value)

    @property
    def kJmolK(self):
        return self.si / 1000

    @kJmolK.setter
    def kJmolK(self, value):
        self.si = to_float(value * 1000)

    def __mul__(self, other):
        if isinstance(other, Temperature):
            si = self.si * other.si
            return Energy(si=si)
        else:
            return super().__mul__(other)

    def __rmul__(self, other):
        if isinstance(other, Temperature):
            si = other.si * self.si
            return Energy(si=si)
        else:
            return super().__rmul__(other)



from .species import Species
from ..quantities import *


class C2H4(Species):
    def __init__(self, T, stoich_coeff=None):
        self.min_temp = Temperature(K=298)
        self.max_temp = Temperature(K=1200)
        self.Hf298 = Energy(kJmol=52.4)
        self.thermo_params = {
            "A": -6.387880,
            "B": 184.4019,
            "C": -112.9718,
            "D": 28.49593,
            "E": 0.315540,
            "F": 48.17332,
            "G": 163.1568,
            "H": 52.46694
        }
        super().__init__(T, stoich_coeff)



import numpy as np


class Quantity:
    def __init__(self, **kwargs):
        if kwargs:
            try:
                key = list(kwargs.keys())[0]
                value = list(kwargs.values())[0]
                if isinstance(value, (list, tuple)):
                    value = np.array(value)
                setattr(self, key, value)
            except:
                raise ValueError("Invalid argument.")

    @property
    def initial(self):
        if isinstance(self.si, np.ndarray):
            return type(self)(si=[col[0] for col in self.si])
        return type(self)(si=self.si[0])
        
    @property
    def final(self):
        if isinstance(self.si, np.ndarray):
            return type(self)(si=[col[-1] for col in self.si])
        return type(self)(si=self.si[-1])
    
    @property
    def size(self):
        if isinstance(self.si, np.ndarray):
            return self.si.size
        return 1
    
    def __getitem__(self, index):
        return type(self)(si=self.si[index])

    def __add__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = self.si + other.si
        return type(self)(si=si)

    def __radd__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = other.si + self.si
        return type(self)(si=si)

    def __sub__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = self.si - other.si
        return type(self)(si=si)

    def __rsub__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = other.si - self.si
        return type(self)(si=si)

    def __mul__(self, other):
        from .unitless import Unitless

        if not isinstance(other, Quantity):
            other = Unitless(si=other)
            # other = Quantity(si=other)
        si = self.si * other.si
        if isinstance(other, Unitless):
            return type(self)(si=si)
        if isinstance(self, Unitless):
            return type(other)(si=si)
        return Quantity(si=si)

    def __rmul__(self, other):
        from .unitless import Unitless

        if not isinstance(other, Quantity):
            other = Unitless(si=other)
            # other = Quantity(si=other)
        si = other.si * self.si
        if isinstance(other, Unitless):
            return type(self)(si=si)
        if isinstance(self, Unitless):
            return type(other)(si=si)
        return Quantity(si=si)

    def __truediv__(self, other):
        from .unitless import Unitless

        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = np.divide(self.si, other.si)
        if isinstance(other, Unitless):
            return type(self)(si=si)
        if type(other) is type(self) and not type(self) is Quantity:
            return Unitless(si=si)
        return Quantity(si=si)

    def __rtruediv__(self, other):
        from .unitless import Unitless

        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = np.divide(other.si, self.si)
        if isinstance(other, Unitless):
            return type(self)(si=si)
        if type(other) is type(self):
            return Unitless(si=si)
        return Quantity(si=si)

    def __pow__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = self.si**other.si
        return Quantity(si=si)

    def __rpow__(self, other):
        if not isinstance(other, Quantity):
            other = Quantity(si=other)
        si = other.si**self.si
        return Quantity(si=si)

    def __neg__(self):
        return type(self)(si=-self.si)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        from .unitless import Unitless
        if ufunc == np.power:
            base, exp = inputs
            if isinstance(base, Quantity) and isinstance(exp, (int, float, np.ndarray)):
                return Quantity(si=np.power(base.si, exp))
            elif isinstance(exp, Quantity) and isinstance(
                base, (int, float, np.ndarray)
            ):
                return Quantity(si=np.power(base, exp.si))
            elif isinstance(exp, Quantity) and isinstance(base, Quantity):
                return Quantity(si=np.power(base.si, exp.si))
            else:
                raise TypeError("Invalid types for np.power with Quantity.")
        elif ufunc == np.add and method == "reduce":
            si = ufunc.reduce(inputs[0].si, **kwargs)
            return type(self)(si=si)
        elif ufunc == np.multiply and method == "reduce":
            si = ufunc.reduce(inputs[0].si, **kwargs)
            return Quantity(si=si)
        elif ufunc == np.exp:
            si = ufunc(inputs[0].si)
            if isinstance(inputs[0], Unitless):
                return Unitless(si=si)
            return Quantity(si=si)
        elif ufunc == np.log:
            si = ufunc(inputs[0].si)
            if isinstance(inputs[0], Unitless):
                return Unitless(si=si)
            return Quantity(si=si)
        return NotImplemented



from .quantity import Quantity
from catrxneng.utils import *


class InversePressure(Quantity):

    @property
    def inv_Pa(self):
        return self.si

    @inv_Pa.setter
    def inv_Pa(self, value):
        self.si = to_float(value)

    @property
    def inv_bar(self):
        return self.si * 100000

    @inv_bar.setter
    def inv_bar(self, value):
        self.si = to_float(value) / 100000

    @property
    def inv_kPa(self):
        return self.si * 1000 

    @inv_kPa.setter
    def inv_kPa(self, value):
        self.si = to_float(value) / 1000

    @property
    def inv_atm(self):
        return self.si * 101325 

    @inv_atm.setter
    def inv_atm(self, value):
        self.si = to_float(value) / 101325

    @property
    def inv_MPa(self):
        return self.si * 1000000 

    @inv_MPa.setter
    def inv_MPa(self, value):
        self.si = to_float(value) / 1000000


from .quantity import Quantity
from catrxneng.utils import *


class VolumetricFlowRate(Quantity):

    @property
    def m3s(self):
        return self.si

    @m3s.setter
    def m3s(self, value):
        self.si = to_float(value)

    @property
    def mLs(self):
        return self.si * 1000 * 1000

    @mLs.setter
    def mLs(self, value):
        self.si = to_float(value) / 1000 / 1000


from .quantity import Quantity
from catrxneng.utils import *


class Mass(Quantity):

    @property
    def kg(self):
        return self.si

    @kg.setter
    def kg(self, value):
        self.si = to_float(value)

    @property
    def g(self):
        return self.si * 1000

    @g.setter
    def g(self, value):
        self.si = to_float(value) / 1000

    def __mul__(self, other):
        from .ghsv import GHSV 
        from .molar_flow_rate import MolarFlowRate
        from .reaction_rate import ReactionRate

        if isinstance(other, (GHSV, ReactionRate)):
            si = self.si * other.si
            return MolarFlowRate(si=si)
        return super().__mul__(other)

    def __rmul__(self, other):
        from .ghsv import GHSV 
        from .molar_flow_rate import MolarFlowRate
        from .reaction_rate import ReactionRate

        if isinstance(other, (GHSV, ReactionRate)):
            si = self.si * other.si
            return MolarFlowRate(si=si)
        return super().__rmul__(other)



from .quantity import Quantity
from catrxneng.utils import *


class Temperature(Quantity):

    @property
    def K(self):
        return self.si

    @K.setter
    def K(self, value):
        self.si = to_float(value)

    @property
    def C(self):
        return self.si - 273

    @C.setter
    def C(self, value):
        self.si = to_float(value + 273)

    def __mul__(self, other):
        from .energy import Energy
        from .entropy import Entropy
        if isinstance(other, Entropy):
            si = self.si * other.si
            return Energy(si=si)
        else:
            return super().__mul__(other)

    def __rmul__(self, other):
        from .energy import Energy
        from .entropy import Entropy
        if isinstance(other, Entropy):
            si = other.si * self.si
            return Energy(si=si)
        else:
            return super().__rmul__(other)



from .reaction import Reaction
from .. import species


class AmmoniaSynthesis(Reaction):
    def __init__(self, T):
        self.components = {
            "N2": species.N2(T=T, stoich_coeff=-0.5),
            "H2": species.H2(T=T, stoich_coeff=-1.5),
            "NH3": species.NH3(T=T, stoich_coeff=1),
            "inert": species.Ar(T=T, stoich_coeff=0),
        }
        super().__init__(T)



import numpy as np
from catrxneng.utils import k
from catrxneng.reactions import RWGS
from catrxneng.quantities import *


class Brubach2022:
    def __init__(self):
        self.components = ("CO2", "H2", "CO", "H2O", "Ar")
        self.Ea_rwgs = Energy(kJmol=115)
        self.Ea_ft = Energy(kJmol=67.8)
        self.a_rwgs = Unitless(si=16.3)
        self.a_ft = Unitless(si=9.07)
        self.b_ft = InversePressure(inv_bar=2.44)
        self.kref_rwgs = RateConstant(order=1.5, molhgcatbar=8.13e-2)
        self.kref_ft = RateConstant(order=2, molhgcatbar=6.39e-2)
        self.Tref = Temperature(C=300)
        self.construct_rate_eqn()

    def k_rwgs(self, T):
        return k(T=T, Ea=self.Ea_rwgs, kref=self.kref_rwgs, Tref=self.Tref)

    def k_ft(self, T):
        return k(T=T, Ea=self.Ea_ft, kref=self.kref_ft, Tref=self.Tref)

    def r_rwgs(self, T, p):
        fwd = p[0] * p[1] ** 0.5
        rev = p[2] * p[3] / (RWGS(T).Keq * p[1] ** 0.5)
        num = self.k_rwgs(T) * (fwd - rev)
        denom = (1 + self.a_rwgs * p[3] / p[1]) ** 2
        rate = num / denom
        return ReactionRate(si=rate.si)

    def r_ft(self, T, p):
        num = self.k_ft(T) * p[1] * p[2]
        denom = (1 + self.a_ft * p[3] / p[1] + self.b_ft * p[2]) ** 2
        rate = num / denom
        return ReactionRate(si=rate.si)

    def construct_rate_eqn(self):
        # def k_rwgs(T):
        #     return k(T=T, Ea=self.Ea_rwgs, kref=self.kref_rwgs, Tref=self.Tref)

        # def k_ft(T):
        #     return k(T=T, Ea=self.Ea_ft, kref=self.kref_ft, Tref=self.Tref)

        # def r_rwgs(T, p):
        #     fwd = p[0] * p[1] ** 0.5
        #     rev = p[2] * p[3] / (RWGS(T).Keq * p[1] ** 0.5)
        #     num = k_rwgs(T) * (fwd - rev)
        #     denom = (1 + self.a_rwgs * p[3] / p[1]) ** 2
        #     rate = num / denom
        #     return ReactionRate(si=rate.si)

        # def r_ft(T, p):
        #     num = k_ft(T) * p[1] * p[2]
        #     denom = (1 + self.a_ft * p[3] / p[1] + self.b_ft * p[2]) ** 2
        #     rate = num / denom
        #     return ReactionRate(si=rate.si)

        self.rate = np.array(
            [
                lambda T, p: -self.r_rwgs(T, p),  # CO2
                lambda T, p: -self.r_rwgs(T, p) - 2 * self.r_ft(T, p),  # H2
                lambda T, p: self.r_rwgs(T, p) - self.r_ft(T, p),  # CO
                lambda T, p: self.r_rwgs(T, p) + self.r_ft(T, p),  # H2O
                lambda T, p: ReactionRate(si=0),  # Ar 
            ]
        )



import numpy as np

from .reaction import Reaction
from ..quantities import *
from ..species import *


class RWGS(Reaction):
    def __init__(self, T):
        # self.components = {
        #     "CO2": CO2(T=T, stoich_coeff=-1),
        #     "H2": H2(T=T, stoich_coeff=-1),
        #     "CO": CO(T=T, stoich_coeff=1),
        #     "H2O": H2O(T=T, stoich_coeff=1),
        #     "inert": Ar(T=T, stoich_coeff=0)
        # }
        self.components = (CO2(T), H2(T), CO(T), H2O(T), Ar(T))
        self.stoich_coeff = Unitless(si=[-1.0, -1.0, 1.0, 1.0, 0.0])
        super().__init__(T)



from scipy.optimize import minimize, fsolve 
from .reaction import Reaction
from ..species import CO2, H2, CH4, H2O


class Sabatier(Reaction):
    def __init__(self, T):
        self.T = T
        self.components = {
            "CO2": CO2(T=T, stoich_coeff=-1),
            "H2": H2(T=T, stoich_coeff=-4),
            "CH4": CH4(T=T, stoich_coeff=1),
            "H2O": H2O(T=T, stoich_coeff=2),
        }
